# type: ignore

from typing import ClassVar, Optional

import pandas as pd
from pydantic import Field
import pytest

from pydantic import ValidationError

from neontology.basenode import BaseNode
from neontology.baserelationship import BaseRelationship


class PracticeNode(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "PracticeNode"
    pp: str


class PracticeRelationship(BaseRelationship):
    source: PracticeNode
    target: PracticeNode
    __relationshiptype__: ClassVar[Optional[str]] = "PRACTICE_RELATIONSHIP"


def test_base_relationship():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    br = PracticeRelationship(source=source_node, target=target_node)

    assert br.source.pp == "Source Node"
    assert br.target.pp == "Target Node"

    assert br.get_relationship_type() == "PRACTICE_RELATIONSHIP"


def test_source_target_type():
    source_node = "Not a BaseNode"
    target_node = "Also not a BaseNode"

    with pytest.raises(ValidationError) as exception_info:
        PracticeRelationship(source=source_node, target=target_node)

    expected = "2 validation errors for PracticeRelationship"

    assert expected in str(exception_info.value)


def test_merge_relationship(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    br = PracticeRelationship(source=source_node, target=target_node)
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "PRACTICE_RELATIONSHIP"


def test_merge_relationship_merge_on_match(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    class TestRel(PracticeRelationship):
        __relationshiptype__: ClassVar[Optional[str]] = "TEST_REL"
        prop_to_merge_on: str = Field(json_schema_extra={"merge_on": True})
        my_prop: str

    br = TestRel(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Foo",
    )
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "TEST_REL"
    assert result[0]["prop_to_merge_on"] == "MergeMe"
    assert result[0]["my_prop"] == "Foo"

    br2 = TestRel(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Bar",
    )
    br2.merge()

    result2 = use_graph.evaluate(cypher)

    assert len(result2) == 1

    assert result2[0]["type_r"] == "TEST_REL"
    assert result2[0]["prop_to_merge_on"] == "MergeMe"
    assert result2[0]["my_prop"] == "Bar"


def test_merge_relationship_merge_on_create(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    class TestRel(PracticeRelationship):
        __relationshiptype__: ClassVar[Optional[str]] = "TEST_REL"
        prop_to_merge_on: str = Field(json_schema_extra={"merge_on": True})
        my_prop: str

    br = TestRel(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Foo",
    )
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "TEST_REL"
    assert result[0]["prop_to_merge_on"] == "MergeMe"
    assert result[0]["my_prop"] == "Foo"

    br2 = TestRel(
        source=source_node,
        target=target_node,
        prop_to_merge_on="Don'tMergeMe",
        my_prop="Bar",
    )
    br2.merge()

    result2 = use_graph.evaluate(cypher)

    assert len(result2) == 2

    my_props = []

    for res in result2:
        my_props.append(res["my_prop"])

    assert set(my_props) == {"Foo", "Bar"}


def test_default_relationship_type():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    br = PracticeRelationship(source=source_node, target=target_node)

    assert br.get_relationship_type() == "PRACTICE_RELATIONSHIP"


def test_default_relationship_type_inherited():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode
        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rel = NewRelType(source=source_node, target=target_node)

    assert rel.get_relationship_type() == "NEWRELTYPE"


def test_defined_relationship_type_inherited():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode
        __relationshiptype__: ClassVar[Optional[str]] = "TEST_RELATIONSHIP_TYPE"

    rel = NewRelType(source=source_node, target=target_node)

    assert rel.get_relationship_type() == "TEST_RELATIONSHIP_TYPE"


def test_merge_relationships_defined_types(use_graph):
    node1 = PracticeNode(pp="Source Node")
    node1.create()

    node2 = PracticeNode(pp="Target Node")
    node2.create()

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode
        __relationshiptype__: ClassVar[Optional[str]] = "TEST_RELATIONSHIP_TYPE"

    rel1 = NewRelType(source=node1, target=node2)

    rel2 = NewRelType(source=node2, target=node1)

    NewRelType.merge_relationships([rel1, rel2])

    cypher = """
    MATCH (src:PracticeNode)-[r]->(tgt:PracticeNode)
    WITH DISTINCT r, TYPE(r) as type_r
    RETURN COLLECT(r{.*, type_r})
    """

    results = use_graph.evaluate(cypher)

    assert len(results) == 2

    for result in results:
        assert result["type_r"] == "TEST_RELATIONSHIP_TYPE"


def test_merge_df(use_graph):
    class SubclassNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "SubclassNode"
        myprop: str

    source_node = SubclassNode(pp="Source Node", myprop="Some Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode

        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rel_records = [{"source": "Source Node", "target": "Target Node"}]

    df = pd.DataFrame.from_records(rel_records)

    NewRelType.merge_df(df, SubclassNode, PracticeNode)

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "NEWRELTYPE"


def test_merge_df_alt_prop(use_graph):
    class SubclassNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "SubclassNode"
        myprop: str

    source_node = SubclassNode(pp="Source Node", myprop="My Prop Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode

        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rel_records = [{"source": "My Prop Value", "target": "Target Node"}]

    df = pd.DataFrame.from_records(rel_records)

    NewRelType.merge_df(df, SubclassNode, PracticeNode, source_prop="myprop")

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "NEWRELTYPE"


def test_merge_df_bad_node_type(use_graph):
    class DifferentNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "DifferentNode"
        __primaryproperty__: ClassVar[str] = "myprop"
        myprop: str

    source_node = PracticeNode(pp="something else")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode

        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rels = [NewRelType(source=source_node, target=target_node)]

    with pytest.raises(TypeError):
        NewRelType.merge_relationships(rels, PracticeNode, DifferentNode)


def test_merge_df_bad_node_type2(use_graph):
    class DifferentNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "DifferentNode"
        __primaryproperty__: ClassVar[str] = "myprop"
        myprop: str

    source_node = DifferentNode(myprop="something", pp="something else")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode

        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rels = [NewRelType(source=source_node, target=target_node)]

    with pytest.raises(TypeError):
        NewRelType.merge_relationships(rels, PracticeNode, PracticeNode)


def test_merge_empty_df():
    df = pd.DataFrame()

    result = PracticeRelationship.merge_df(df)

    assert result is None


def test_merge_records(use_graph):
    class SubclassNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "SubclassNode"
        myprop: str

    source_node = SubclassNode(pp="Source Node", myprop="My Prop Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    class NewRelType(BaseRelationship):
        source: SubclassNode
        target: PracticeNode

        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    records = [{"source": "Source Node", "target": "Target Node"}]

    NewRelType.merge_records(records)

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    WITH r, TYPE(r) as type_r
    RETURN COLLECT(DISTINCT r{.*, type_r})
    """

    result = use_graph.evaluate(cypher)

    assert len(result) == 1

    assert result[0]["type_r"] == "NEWRELTYPE"



# flake8: noqa

from .basenode import BaseNode
from .baserelationship import BaseRelationship
from .graphconnection import GraphConnection, init_neontology
from .utils import auto_constrain

__all__ = [
    # BaseNode
    "BaseNode",
    # BaseRelationship
    "BaseRelationship",
    # GraphConnection
    "init_neontology",
    "GraphConnection",
    # utils
    "auto_constrain",
]



# type: ignore
from typing import ClassVar, Optional
from datetime import datetime
from uuid import UUID

import pandas as pd
from pydantic import Field, field_validator
import pytest

from neontology import BaseNode


class PracticeNode(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "PracticeNode"

    pp: str


def test_set_pp_field_valid():
    tn = PracticeNode(pp="Some Value")

    assert tn.pp == "Some Value"


def test_create(use_graph):
    tn = PracticeNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:PracticeNode)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate(cypher)

    assert result.has_label("PracticeNode")

    assert result.get("pp") == "Test Node"


def test_no_primary_label():
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    with pytest.raises(AttributeError):
        SpecialPracticeNode(pp="Test Node")


def test_none_primary_label():
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = None
        pp: str

    with pytest.raises(NotImplementedError):
        SpecialPracticeNode(pp="Test Node")


def test_create_multilabel(use_graph):
    class MultipleLabelNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "PrimaryLabel"
        __secondarylabels__: ClassVar[Optional[list]] = ["ExtraLabel1", "ExtraLabel2"]
        pp: str

    tn = MultipleLabelNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:ExtraLabel1)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate(cypher)

    assert result.has_label("PrimaryLabel")
    assert result.has_label("ExtraLabel1")
    assert result.has_label("ExtraLabel2")

    assert result.get("pp") == "Test Node"


def test_create_multilabel_inheritance(use_graph):
    class Mammal(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __secondarylabels__: ClassVar[Optional[list]] = ["Mammal"]
        pp: str

    class Human(Mammal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Human"
        pp: str

    tn = Human(pp="Bob")

    tn.create()

    cypher = """
    MATCH (n:Human)
    WHERE n.pp = 'Bob'
    RETURN n
    """

    result = use_graph.evaluate(cypher)

    assert result.has_label("Human")
    assert result.has_label("Mammal")

    assert result.get("pp") == "Bob"


def test_create_multilabel_inheritance_multiple(use_graph):
    class Animal(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __secondarylabels__: ClassVar[Optional[list]] = ["Animal"]
        pp: str

    class Human(Animal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Human"
        pp: str

    class Elephant(Animal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Elephant"
        pp: str

    human = Human(pp="Bob")

    human.create()

    elephant = Elephant(pp="Bob")

    # this should create a new node because the primary label is different
    elephant.create()

    elephant2 = Elephant(pp="Bob")

    # this shouldn't create a new node because the primary label is the same
    elephant2.merge()

    cypher = """
    MATCH (n:Animal)
    WHERE n.pp = 'Bob'
    RETURN COUNT(DISTINCT n)
    """

    result = use_graph.evaluate(cypher)

    assert result == 2


def test_merge_defined_label_inherited(use_graph):
    class SpecialPracticeNode(BaseNode):
        __primarylabel__: ClassVar[Optional[str]] = "SpecialTestLabel"
        __primaryproperty__: ClassVar[str] = "pp"

        pp: str

    tn = SpecialPracticeNode(pp="Special Test Node")

    tn.merge()

    cypher = """
    MATCH (n:SpecialTestLabel)
    WHERE n.pp = 'Special Test Node'
    RETURN n
    """

    result = use_graph.evaluate(cypher)

    assert result.has_label("SpecialTestLabel")

    assert result.get("pp") == "Special Test Node"


def test_merge_multiple_defined_label_inherited(use_graph):
    class SpecialPracticeNode(BaseNode):
        __primarylabel__: ClassVar[Optional[str]] = "SpecialTestLabel"
        __primaryproperty__: ClassVar[str] = "pp"

        pp: str

    tn = SpecialPracticeNode(pp="Special Test Node")

    tn2 = SpecialPracticeNode(pp="Special Test Node2")

    SpecialPracticeNode.merge_nodes([tn, tn2])

    cypher = """
    MATCH (n:SpecialTestLabel)
    RETURN COLLECT(DISTINCT n.pp) as node_names
    """

    results = use_graph.evaluate(cypher)

    assert "Special Test Node" in results
    assert "Special Test Node2" in results


def test_create_multiple_defined_label_inherited(use_graph):
    class SpecialPracticeNode(BaseNode):
        __primarylabel__: ClassVar[Optional[str]] = "SpecialTestLabel"
        __primaryproperty__: ClassVar[str] = "pp"

        pp: str

    tn = SpecialPracticeNode(pp="Special Test Node")

    tn2 = SpecialPracticeNode(pp="Special Test Node2")

    SpecialPracticeNode.create_nodes([tn, tn2])

    cypher = """
    MATCH (n:SpecialTestLabel)
    RETURN COLLECT(DISTINCT n.pp) as node_names
    """

    results = use_graph.evaluate(cypher)

    assert "Special Test Node" in results
    assert "Special Test Node2" in results


def test_creation_datetime(use_graph):
    """
    Check we can manually define the created datetime, and then check we can
    query for it using neo4j DateTime type.
    """

    my_datetime = datetime(year=2022, month=5, day=4, hour=3, minute=21)

    bn = PracticeNode(pp="Test Node", created=my_datetime)

    bn.create()

    cypher = """
    MATCH (n:PracticeNode)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate(cypher)

    # if the created date has been stored as a Neo4j datetime,
    # it should come back as an interchange DateTime format from py2neo
    # we call to_native to convert it back to a native Python datetime
    assert result.get("created").to_native() == my_datetime


def test_match_nodes(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn2 = PracticeNode(pp="Special Test Node2")

    PracticeNode.merge_nodes([tn, tn2])

    results = PracticeNode.match_nodes()

    for result in results:
        assert isinstance(result, PracticeNode)

    pps = [x.pp for x in results]

    assert "Special Test Node" in pps

    assert "Special Test Node2" in pps


def test_match_node(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn.create()

    result = PracticeNode.match("Special Test Node")

    assert isinstance(result, PracticeNode)

    assert "Special Test Node" == result.pp


def test_match_none(use_graph):
    bn = PracticeNode(pp="Test Node")
    bn.create()

    result = PracticeNode.match("Does Not Exist")

    assert result is None


def test_match_many_none(use_graph):
    class PracticeNode2(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[str] = "PracticeNode2"

        pp: str

    bn = PracticeNode2(pp="Test Node")
    bn.create()

    results = PracticeNode.match_nodes()

    assert results == []


def test_delete_node(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn.create()

    result = PracticeNode.match("Special Test Node")

    assert isinstance(result, PracticeNode)

    PracticeNode.delete("Special Test Node")

    result = PracticeNode.match("Special Test Node")

    assert result is None


@pytest.mark.parametrize(
    "field_type,python_value,neo4j_values",
    [
        (str, "hello world", ["hello world"]),
        (tuple, ("hello", "world"), [["hello", "world"]]),
        (set, {"foo", "bar"}, [["bar", "foo"], ["foo", "bar"]]),
        (
            UUID,
            UUID("32d4a4cb-29c3-4aa8-9b55-7790431819e3"),
            ["32d4a4cb-29c3-4aa8-9b55-7790431819e3"],
        ),
        (
            datetime,
            datetime(year=1984, month=1, day=2),
            [datetime(year=1984, month=1, day=2)],
        ),
        (
            list,
            ["foo", "bar"],
            [["foo", "bar"]],
        ),
        (
            list,
            [1, 2, 3],
            [[1, 2, 3]],
        ),
    ],
)
def test_neo4j_dict_create(use_graph, field_type, python_value, neo4j_values):
    class TestModel(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel"
        pp: str
        test_prop: field_type

    pp = "test_node"

    testmodel = TestModel(test_prop=python_value, pp=pp)

    testmodel.create()

    result = TestModel.match(pp)

    assert result.test_prop == python_value

    cypher = """
    MATCH (n:TestModel)
    WHERE n.pp = 'test_node'
    RETURN n
    """

    cypher_result = use_graph.evaluate(cypher)

    assert cypher_result.get("test_prop") in neo4j_values


def test_set_on_match(use_graph):
    """Check that we successfully identify field to set on match"""

    class TestModel(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel"
        pp: str = "test_node"
        only_set_on_match: Optional[str] = Field(
            json_schema_extra={"set_on_match": True}, default=None
        )
        normal_field: str

    test_node = TestModel(only_set_on_match="Foo", normal_field="Bar", pp="test_node")
    test_node.merge()

    cypher = """
    MATCH (n:TestModel)
    WHERE n.pp = 'test_node'
    RETURN n
    """

    cypher_result = use_graph.evaluate(cypher)

    assert cypher_result.get("only_set_on_match") is None
    assert cypher_result.get("normal_field") == "Bar"

    test_node2 = TestModel(only_set_on_match="Foo", normal_field="Bar", pp="test_node")
    test_node2.merge()

    cypher_result2 = use_graph.evaluate(cypher)

    assert cypher_result2.get("only_set_on_match") == "Foo"


def test_set_on_create(use_graph):
    """Check that we successfully identify field to set on match"""

    class TestModel(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel"
        pp: str = "test_node"
        only_set_on_create: str = Field(json_schema_extra={"set_on_create": True})
        normal_field: str

    test_node = TestModel(only_set_on_create="Foo", normal_field="Bar", pp="test_node")
    test_node.merge()

    cypher = """
    MATCH (n:TestModel)
    WHERE n.pp = 'test_node'
    RETURN n
    """

    cypher_result = use_graph.evaluate(cypher)

    assert cypher_result.get("only_set_on_create") == "Foo"
    assert cypher_result.get("normal_field") == "Bar"

    test_node2 = TestModel(only_set_on_create="Fee", normal_field="Fi", pp="test_node")
    test_node2.merge()

    cypher_result2 = use_graph.evaluate(cypher)

    assert cypher_result2.get("only_set_on_create") == "Foo"
    assert cypher_result2.get("normal_field") == "Fi"


def test_merge_df_with_duplicates(use_graph):
    class Person(BaseNode):
        __primaryproperty__: ClassVar[str] = "identifier"
        __primarylabel__: ClassVar[
            str
        ] = "PersonLabel"  # optionally specify the label to use

        name: str
        age: int
        identifier: Optional[str] = Field(default=None, validate_default=True)

        @field_validator("identifier")
        def set_identifier(cls, v, values):
            if v is None:
                v = f"{values.data['name']}_{values.data['age']}"

            return v

    people_records = [
        {"name": "arthur", "age": 70},
        {"name": "betty", "age": 65},
        {"name": "betty", "age": 65},
        {"name": "ted", "age": 50},
        {"name": "betty", "age": 75},
        {"name": "arthur", "age": 70},
    ]

    people_df = pd.DataFrame.from_records(people_records)

    results = Person.merge_df(people_df)

    names = [x.name for x in results]

    assert names == ["arthur", "betty", "betty", "ted", "betty", "arthur"]

    assert results[0].identifier == results[5].identifier

    assert results[1].identifier == results[2].identifier

    assert results[1].identifier != results[4].identifier

    assert results[3].name == "ted"


def test_merge_empty_df():
    df = pd.DataFrame()

    result = PracticeNode.merge_df(df)

    assert len(result) == 0
    assert isinstance(result, pd.Series)



import pytest

from neontology.graphconnection import GraphConnection


def test_evaluate_query_single(use_graph):
    gc = GraphConnection()

    create_cypher = """
    CREATE (tn:TestNode {name: "Foo Bar"})
    RETURN tn.name
    """

    result = gc.evaluate_query_single(create_cypher)

    assert result == "Foo Bar"


def test_evaluate_query_node(use_graph):
    gc = GraphConnection()

    create_cypher = """
    CREATE (tn:TestNode {name: "Foo Bar"})
    RETURN tn
    """

    result = gc.evaluate_query_single(create_cypher)

    assert dict(result)["name"] == "Foo Bar"


def test_evaluate_query_single_multiple(use_graph):
    gc = GraphConnection()

    create_cypher = """
    CREATE (tn1:TestNode {name: "Foo"})
    CREATE (tn2:TestNode {name: "Bar"})
    """

    gc.evaluate_query_single(create_cypher)

    match_cypher = """
    MATCH (n:TestNode)
    RETURN n
    """

    with pytest.warns(UserWarning):
        gc.evaluate_query_single(match_cypher)


def test_evaluate_query_single_collected(use_graph):
    gc = GraphConnection()

    create_cypher = """
    CREATE (tn1:TestNode {name: "Foo"})
    CREATE (tn2:TestNode {name: "Bar"})
    """

    result = gc.evaluate_query_single(create_cypher)

    match_cypher = """
    MATCH (n:TestNode)
    RETURN COLLECT(n.name)
    """

    result = gc.evaluate_query_single(match_cypher)

    assert result == ["Foo", "Bar"]
